"""
Provides information about archive

Examples:
 - What was said about Musk
  $ --search "title=*Musk*"
 - What was said about Musk (title, link, description, etc)
  $ --search "Musk"

TODO
 - Output formats? (md)?
 - Maybe it could produce a chart?

"""

import argparse
import time
import os
import json
from sqlalchemy import create_engine

from .utils.omnisearch import SingleSymbolEvaluator, EquationEvaluator, OmniSearch
from .utils.alchemysearch import (
    AlchemySymbolEvaluator,
    AlchemyEquationEvaluator,
    AlchemySearch,
)
from .utils.reflected import (
    ReflectedEntryTable,
    ReflectedUserTags,
    ReflectedSocialData,
)


def print_time_diff(start_time):
    elapsed_time_seconds = time.time() - start_time
    elapsed_minutes = int(elapsed_time_seconds // 60)
    elapsed_seconds = int(elapsed_time_seconds % 60)
    print(f"Time: {elapsed_minutes}:{elapsed_seconds}")


class RowHandler(object):

    def __init__(self, parser=None, engine=None, connection=None):
        self.parser = parser
        self.start_time = time.time()
        self.engine = engine
        self.connection = connection

        self.files = []

        self.total_entries = 0
        self.good_entries = 0
        self.dead_entries = 0

    def print_entry(self, entry):
        level = self.parser.get_verbosity_level()

        text = ""

        if self.parser.args.description:
            print("---------------------")

        text = "[{:03d}] {}".format(entry.page_rating_votes, entry.link)

        if self.parser.args.title:
            if entry.title:
                text += " " + entry.title

        if self.parser.args.source:
            source_id = entry.source
            if source_id:
                r = ReflectedEntryTable(self.engine, self.connection)
                source = r.get_source(source_id)
                text += " [{}]".format(source.title)

        print(text)

        if self.parser.args.date_published:
            date_published = entry.date_published
            if date_published:
                print(date_published)

        if self.parser.args.description:
            description = entry.description
            if description:
                print(description)

        if self.parser.args.tags:
            tags_table = ReflectedUserTags(self.engine, self.connection)
            tags = tags_table.get_tags_string(entry.id)
            if tags and tags != "":
                self.print_tags(tags)

        if self.parser.args.social:
            social_table = ReflectedSocialData(self.engine, self.connection)
            social = social_table.get(entry.id)
            if social is not None:
                self.print_social(social)

        if self.parser.args.status:
            print(entry.status_code)

    def print_tags(self, tags):
        print(tags)

    def print_social(self, social):
        if (
            social.view_count is not None
            and social.thumbs_up is not None
            and social.thumbs_down is not None
        ):
            print(
                f"V:{social.view_count} TU:{social.thumbs_up} TD:{social.thumbs_down}"
            )
        else:
            if social.view_count:
                print(f"F:{social.view_count}")

            if social.thumbs_up:
                print(f"F:{social.thumbs_up}")

            if social.thumbs_down:
                print(f"F:{social.thumbs_down}")

            if social.upvote_diff:
                print(f"S:{social.upvote_diff}")

            if social.upvote_ratio:
                print(f"S:{social.upvote_ratio}")

            if social.followers_count:
                print(f"F:{social.followers_count}")

            if social.stars:
                print(f"S:{social.stars}")

    def get_time_diff(self):
        return time.time() - self.start_time

    def handle_row(self, row):
        """
        Row is to be expected a 'dict', eg. row["link"]
        """
        link = row.link

        level = self.parser.get_verbosity_level()

        self.print_entry(row)

        self.total_entries += 1

    def summary(self):
        if self.parser.args.summary:
            if self.parser.args.verify:
                print(
                    "total:{} good:{} dead:{}".format(
                        self.total_entries, self.good_entries, self.dead_entries
                    )
                )
            else:
                print("total:{}".format(self.total_entries))


class DbAnalyzer(object):
    def __init__(self, input_db, parser=None):
        self.parser = parser
        self.result = None
        self.engine = None
        self.input_db = input_db

    def print_summary(self, print_columns=False):
        db = self.input_db

        if not os.path.isfile(db):
            print("File does not exist:{}".format(db))
            return

        self.engine = create_engine("sqlite:///" + db)
        with self.engine.connect() as connection:
            r = ReflectedEntryTable(self.engine, connection)
            r.print_summary(print_columns)

    def search(self):
        if self.is_db_scan():
            file = self.input_db
            if not os.path.isfile(file):
                print("File does not exist:{}".format(file))
                return

            print("Creating engine")
            self.engine = create_engine("sqlite:///" + self.input_db)
            print("Creating engine DONE")

            with self.engine.connect() as connection:
                self.connection = connection

                row_handler = RowHandler(self.parser, self.engine, self.connection)

                print("Starting alchemy")
                searcher = AlchemySearch(
                    self.engine,
                    self.parser.args.search,
                    row_handler=row_handler,
                    args=self.parser.args,
                    connection=self.connection,
                )
                print("Starting alchemy DONE")

                print("Searching...")
                searcher.search()

    def is_db_scan(self):
        if self.parser.args.db:
            return True

        return False


class Parser(object):

    def parse(self):
        self.parser = argparse.ArgumentParser(description="Data analyzer program")
        self.parser.add_argument("--db", help="DB to be scanned")

        self.parser.add_argument(
            "--search", help="Search, with syntax same as the main program / site."
        )
        self.parser.add_argument(
            "--order-by", default="page_rating_votes", help="order by column."
        )
        self.parser.add_argument("--asc", action="store_true", help="order ascending")
        self.parser.add_argument("--desc", action="store_true", help="order descending")
        self.parser.add_argument("--table", default="linkdatamodel", help="Table name")

        self.parser.add_argument("--title", action="store_true", help="displays title")
        self.parser.add_argument(
            "--description", action="store_true", help="displays description"
        )
        self.parser.add_argument(
            "--status", action="store_true", help="displays status"
        )
        self.parser.add_argument("--tags", action="store_true", help="displays tags")
        self.parser.add_argument(
            "--social", action="store_true", help="displays social data"
        )
        self.parser.add_argument(
            "--date-published", action="store_true", help="displays date-published"
        )
        self.parser.add_argument(
            "--source", action="store_true", help="displays source"
        )

        self.parser.add_argument(
            "--summary", action="store_true", help="displays summary of tables"
        )
        self.parser.add_argument(
            "--columns",
            action="store_true",
            help="displays summary of tables column nmaes",
        )

        self.parser.add_argument(
            "-i", "--ignore-case", action="store_true", help="Ignores case"
        )
        self.parser.add_argument("-v", "--verbosity", help="Verbosity level")

        self.args = self.parser.parse_args()

        return True

    def get_verbosity_level(self):
        level = 1
        if self.args.verbosity:
            try:
                level = int(self.args.verbosity)
            except Exception as E:
                print(str(E))

        return level


def main():
    p = Parser()
    if not p.parse():
        print("Could not parse options")
        return

    start_time = time.time()

    m = DbAnalyzer(input_db=p.args.db, parser=p)
    if p.args.summary:
        m.print_summary(p.args.columns)
    else:
        m.search()

    print_time_diff(start_time)


if __name__ == "__main__":
    main()



"""
@brief Converts JSON files to SQLite DB

SQLite can easily be imported and used by other projects.
"""

import os
import sqlite3
import json
import argparse
import time

#from .utils.sqlmodel import SqlModel
from dateutil import parser


class DirReader(object):
    def __init__(self, source_files_directory, accepted_extensions=None):
        self.dir = source_files_directory
        if accepted_extensions is None:
            self.accepted_extensions = [".json"]

    def get_files(self):
        file_list = []
        for root, dirs, files in os.walk(self.dir):
            for file in files:
                file_split = os.path.splitext(file)
                if file_split[1] in self.accepted_extensions:
                    file_list.append(os.path.join(root, file))

        file_list = sorted(file_list)
        return file_list


class JSON2Db(object):
    """
    Performs actual conversion between file and database
    """

    def __init__(self, db_conn, parser):
        self.conn = db_conn
        self.parser = parser
        self.file_reader = DirReader(source_files_directory=parser.dir)
        self.files = self.file_reader.get_files()

    def convert(self):
        total_num_files = len(self.files)

        for row, afile in enumerate(self.files):
            print("[{}/{}]: file:{}".format(row, total_num_files, afile))
            self.convert_file(afile)

    def convert_file(self, file_name):
        data = self.read_file(file_name)
        if not data:
            return

        total_rows = len(data)

        for row, entry in enumerate(data):
            if "link" in entry:
                if self.parser and self.parser.preserve_id:
                    if "id" not in entry:
                        print("Entry {} is missing ID".format(entry["link"]))
                        continue
                else:
                    entry["id"] = row

                if self.is_entry_to_be_added(entry):
                    if self.conn.entries_table.add_entry(entry):
                        if self.parser.args.verbose:
                            print(
                                " -> [{}/{}] Link:{} Added".format(
                                    row, total_rows, entry["link"]
                                )
                            )
                    else:
                        print(
                            " -> [{}/{}] Link:{} NOT Added".format(
                                row, total_rows, entry["link"]
                            )
                        )
                else:
                    if self.parser.args.verbose:
                        print(
                            " -> [{}/{}] Link:{} Skipped".format(
                                row, total_rows, entry["link"]
                            )
                        )

    def is_entry_to_be_added(self, entry):
        # entry already exists
        if self.conn.entries_table.is_entry(entry):
            return False

        if self.parser.vote_min:
            if int(entry["page_rating_votes"]) < self.parser.vote_min:
                return False

        return True

    def read_file_contents(self, file_name):
        with open(file_name, "r") as f:
            return f.read()

    def read_file(self, file_name):
        text = self.read_file_contents(file_name)

        try:
            j = json.loads(text)

            if "links" in j:
                return j["links"]
            if "sources" in j:
                return j["sources"]

            return j
        except Exception as e:
            print("Could not read file: {}".format(afile))


class Parser(object):
    def parse(self):
        self.parser = argparse.ArgumentParser(description="Data converter program")
        self.parser.add_argument("--dir", help="Directory to be scanned")
        self.parser.add_argument(
            "--db", default="converted.sqlite", help="Database output file"
        )
        self.parser.add_argument(
            "--preserve-id", action="store_true", help="Preserves ID of objects"
        )
        self.parser.add_argument("--vote-min", help="Minimum amount of entry vote")
        self.parser.add_argument("--language", help="Accept language")  # TODO implement
        self.parser.add_argument("--entries", help="Convert entries")  # TODO implement
        self.parser.add_argument("--sources", help="Convert sources")  # TODO implement
        self.parser.add_argument(
            "--verbose", action="store_true", help="Shows more info"
        )

        self.args = self.parser.parse_args()

        if self.args.dir:
            self.dir = self.args.dir
        else:
            self.dir = None

        if self.args.preserve_id:
            self.preserve_id = self.args.preserve_id
        else:
            self.preserve_id = None

        if self.args.vote_min:
            self.vote_min = int(self.args.vote_min)
        else:
            self.vote_min = None


def main():
    print("Starting processing")
    parser = Parser()
    parser.parse()

    db = SqlModel(database_file=parser.args.db)

    try:
        start_time = time.time()

        c = JSON2Db(db, parser)
        c.convert()

        elapsed_time_seconds = time.time() - start_time
        elapsed_minutes = int(elapsed_time_seconds // 60)
        elapsed_seconds = int(elapsed_time_seconds % 60)
        print(f"Time: {elapsed_minutes}:{elapsed_seconds}")

    except Exception as e:
        print("Exception: {}".format(e))
    except KeyboardInterrupt as e:
        print("Exception: {}".format(e))

    db.close()
    print("Processing DONE")


if __name__ == "__main__":
    main()



"""
Cannot have any dependencies to django
"""

try:
    from sympy import sympify
    import sympy
except Exception as E:
    pass


class SingleSymbolEvaluator(object):
    """
    Evaluates single operation, like "Thing = Anything"
    """

    def __init__(self):
        self.not_translated_conditions = {}
        self.translated_conditions = {}
        self.translatable_names = []
        self.errors = []

    def evaluate_symbol(self, symbol):
        """
        Splits one symbol like "Thing = Anything" into 3 parts

        @returns evaluated symbol value
        """
        condition_data = self.split_symbol(symbol)
        if condition_data:
            return self.evaluate_complex_symbol(symbol, condition_data)
        else:
            return self.evaluate_simple_symbol(symbol)

    def evaluate_complex_symbol(self, symbol, condition_data):
        # print("Condition data {}".format(condition_data))

        if self.is_translatable(condition_data):
            self.enhance_condition_data(condition_data)

            self.translated_conditions[condition_data[0]] = condition_data[2]
            condition_data = self.translate_condition(condition_data)
            return self.evaluate_complex_translated_symbol(symbol, condition_data)
        else:
            self.not_translated_conditions[condition_data[0]] = condition_data[2]
            return self.evaluate_complex_not_translated_symbol(symbol, condition_data)

    def evaluate_complex_translated_symbol(self, symbol, condition_data):
        return condition_data

    def evaluate_complex_not_translated_symbol(self, symbol, condition_data):
        return condition_data

    def evaluate_simple_symbol(self, symbol):
        return symbol

    def enhance_condition_data(self, condition_data):
        return condition_data

    def get_operators(self):
        """
        first 2 char operators, then 1 char operators
        """
        return ["===", "==", "?=", "<=", ">=", "~", "=", "<", ">"]

    def cleanup_left_operator_part(self, left_part):
        return left_part.strip()

    def cleanup_right_operator_part(self, right_part):
        right_part = right_part.strip()

        if right_part.startswith('"') and right_part.endswith('"'):
            right_part = right_part[1:-1]
        if right_part.startswith("'") and right_part.endswith("'"):
            right_part = right_part[1:-1]

        return right_part

    def split_symbol(self, symbol):
        for op in self.get_operators():
            wh = symbol.find(op)
            if wh >= 0:
                sp = [symbol[:wh], symbol[wh + len(op) :]]

                left_part = self.cleanup_left_operator_part(sp[0])
                right_part = self.cleanup_right_operator_part(sp[1])

                return [left_part, op, right_part]

    def is_translatable(self, condition):
        """
        If dev does not provide any translate validation - everything will be translated
        """
        if len(self.translatable_names) == 0:
            return True

        for name in self.translatable_names:
            if condition[0] == name:
                return True

        return False

    def translate_condition(self, condition_data):
        """
        Translate user input into system understood syntax.

        Example for django:
         title == 'something'
        Translated:
         title__iexact : 'something'

        by default - not translate
        """
        return condition_data

    def set_translation_mapping(self, names):
        self.translatable_names = names

    def add_error(self, error):
        self.errors.append(error)


class EquationTranslator(object):
    def __init__(self, data):
        self.data = data
        self.current_symbol = ord("A") - 1

    def get_operators():
        return ("(", ")", "&", "|", "~", "^", "!")

    def get_whitespaces():
        return (" ", "\t")

    def is_operator(char):
        return char in EquationTranslator.get_operators()

    def is_whitespace(char):
        return char in EquationTranslator.get_whitespaces()

    def process(self):
        result_string = ""
        inside_text = False
        self.conditions = {}
        self.current_condition = ""

        for char in self.data:
            if EquationTranslator.is_operator(char):
                inside_text = False
                result_string += char

                self.add_condition()

            elif inside_text:
                self.current_condition += char
            else:
                self.current_condition += char

                if not EquationTranslator.is_whitespace(char):
                    inside_text = True
                    result_string += self.get_next_symbol()

        self.add_condition()

        return result_string, self.conditions

    def add_condition(self):
        self.current_condition = self.current_condition.strip()
        if self.current_condition != "":
            self.conditions[self.get_current_symbol()] = self.current_condition
            self.current_condition = ""

    def get_current_symbol(self):
        return chr(self.current_symbol)

    def get_next_symbol(self):
        self.current_symbol += 1
        return chr(self.current_symbol)


class EquationEvaluator(object):
    """
    Evaluates equations
    """

    def __init__(self, data, symbol_evaluator):
        self.data = data
        self.symbol_evaluator = symbol_evaluator
        self.known_results = {}
        self.expr = None

    def translate_to_symbol_notation(self, data):
        """
        Sympy operates on symbol notation
        """
        eq = EquationTranslator(data)
        self.eq_string, self.conditions = eq.process()

    def process(self):
        if self.expr is None:
            self.translate_to_symbol_notation(self.data)
            self.expr = sympify(self.eq_string)

        return self.process_internal(self.expr)

    def process_internal(self, expr):
        for arg in expr.args:
            self.process_internal(arg)

        if expr.func == sympy.core.symbol.Symbol:
            symbol = str(expr)

            return self.evaluate_symbol(symbol)
        else:
            function = str(expr.func)
            operation_symbol = str(expr)
            # print("Operation: {}".format(function))

            return self.evaluate_function_and_store(
                operation_symbol, function, expr.args
            )

        # print(f'arg {expr}')
        # print(f'arg.func: {expr.func}')
        # print(f'arg.args: {expr.args}')

    def evaluate_symbol(self, symbol):
        condition_text = self.conditions[symbol]
        # print("Evaluation condition {} {}".format(symbol, condition_text))

        self.known_results[symbol] = self.symbol_evaluator.evaluate_symbol(
            condition_text
        )

        return self.known_results[symbol]

    def evaluate_function_and_store(self, operation_symbol, function, args):
        args0 = str(args[0])
        args0 = self.known_results[args0]

        if len(args) > 1:
            args1 = str(args[1])
            args1 = self.known_results[args1]
        else:
            args1 = None

        val = self.evaluate_function(operation_symbol, function, args0, args1)
        self.known_results[operation_symbol] = val
        return self.known_results[operation_symbol]

    def evaluate_function(self, operation_symbol, function, args0, args1):
        if function == "And":  # & sign
            return args0 & args1
        elif function == "Or":  # | sign
            return args0 | args1
        elif function == "Not":  # ~ sign
            return ~args0
        else:
            raise NotImplementedError("Not implemented function: {}".format(function))

    def reevaluate(self):
        self.known_results = {}


class OmniSearch(object):
    def __init__(self, search_query, symbol_evaluator=None, equation_evaluator=None):
        """
        I assume that either symbol_evaluator is specified, or equation_evaluator
        """
        self.search_query = search_query
        self.query_result = None
        self.errors = []

        if symbol_evaluator:
            self.symbol_evaluator = symbol_evaluator
        else:
            self.symbol_evaluator = SingleSymbolEvaluator()

        if equation_evaluator:
            self.equation_evaluator = equation_evaluator
        else:
            self.equation_evaluator = EquationEvaluator(
                self.search_query, self.symbol_evaluator
            )

    def set_symbol_evaluator(self, symbol_evaluator):
        self.symbol_evaluator = symbol_evaluator
        self.equation_evaluator.symbol_evaluator = symbol_evaluator

    def set_translation_mapping(self, name_mapping):
        """
        We allow only certain symbols to be translated in query.
        We operate on white-list.
        """
        self.symbol_evaluator.set_translation_mapping(name_mapping)
        self.equation_evaluator.symbol_evaluator.set_translation_mapping(name_mapping)

    def get_translated_conditions(self):
        return self.symbol_evaluator.translated_conditions

    def get_not_translated_conditions(self):
        return self.symbol_evaluator.not_translated_conditions

    def is_complex_query(self):
        uses_operator = False

        operators = set()
        for symbol in self.symbol_evaluator.get_operators():
            operators.add(symbol)
        for symbol in EquationTranslator.get_operators():
            operators.add(symbol)

        for symbol in operators:
            if self.search_query.find(symbol) >= 0:
                return True

        return False

    def get_query_result(self):
        """
        API
        """
        if self.query_result is not None:
            return self.query_result

        self.query_result = self.get_combined_query()
        return self.query_result

    def get_combined_query(self):
        if self.search_query:
            query_result = self.equation_evaluator.process()
            return query_result

    def reevaluate(self):
        self.query_result = None
        self.equation_evaluator.reevaluate()



"""
Converts Database to information about RSS feeds.

TODO - copy tags from origin to RSS feeds
"""

import shutil
import argparse
from pathlib import Path
from sqlalchemy import create_engine

from webtoolkit import RemoteUrl, BaseUrl
from .utils.reflected import *


class Db2Feeds(object):
    """
    Converter DB -> feeds.
    """

    def __init__(
        self,
        input_db=None,
        verbose=True,
        remote_server="",
        output_db=None,
        output_format=None,
        read_internet_links=False,
        update_rss=False,
    ):
        self.input_db = input_db
        self.verbose = verbose
        self.remote_server = remote_server
        self.update_rss = update_rss
        self.read_internet_links = read_internet_links

        self.output_db = output_db
        self.output_format = output_format

        if self.output_db:
            self.output_format = "SQLITE"
        self.make_output_db()

    def convert(self):
        self.engine = create_engine(f"sqlite:///{self.input_db}")
        with self.engine.connect() as connection:
            self.connection = connection
            if self.new_engine:
                with self.new_engine.connect() as new_connection:
                    self.new_connection = new_connection
                    self.read()
            else:
                self.read()

    def make_output_db(self):
        if self.output_format != "SQLITE":
            self.new_engine = None
            self.new_connection = None
            return

        new_path = Path(self.output_db)
        if new_path.exists():
            new_path.unlink()

        shutil.copy(self.input_db, self.output_db)

        self.new_engine = create_engine(f"sqlite:///{self.output_db}")

    def read(self):
        self.truncate_tables()

        table = ReflectedEntryTable(self.engine, self.connection)

        new_table = None
        if self.new_engine:
            new_table = ReflectedTable(self.new_engine, self.new_connection)
            new_table.vacuum()

        for entry in table.get_entries_good():
            if self.read_internet_links and self.remote_server:
                url = RemoteUrl(remote_server=self.remote_server, url=entry.link)
                feeds = url.get_feeds()
                if len(feeds) == 0:
                    url.get_response()
            else:
                url = BaseUrl(entry.link)

            feeds = url.get_feeds()

            for feed in feeds:
                data = {}
                data["link"] = feed
                data["title"] = entry.title
                data["page_rating_votes"] = entry.page_rating_votes
                data["manual_status_code"] = entry.manual_status_code

                # not null requirement
                data["source_url"] = ""
                data["permanent"] = False
                data["bookmarked"] = False
                data["status_code"] = entry.status_code
                data["contents_type"] = 0
                data["page_rating_contents"] = 0
                data["page_rating_visits"] = 0
                data["page_rating"] = 0

                if self.update_rss and self.remote_server:
                    url_feed = RemoteUrl(remote_server=self.remote_server, url=feed)
                    url_feed.get_response()

                    data["title"] = url_feed.get_title()
                    data["description"] = url_feed.get_description()
                    data["status_code"] = url_feed.get_status_code()

                self.print_data(entry, data)

                if new_table:
                    self.copy_entry(entry, new_table, data)

    def copy_entry(self, entry, table, data):
        """
        TODO copy from origin table tags
        """
        table.insert_json_data("linkdatamodel", data)

        entry_compacted_tags = ReflectedEntryCompactedTags(self.engine, self.connection)
        tags = entry_compacted_tags.get_tags(entry.id)

        entry_tag_data = {}
        for tag in tags:
            entry_tag_data["tag"] = tag
            entry_tag_data["entry_id"] = entry.id
            user_tags.insert_json_data("entrycompactedtags", entry_tag_data)

    def truncate_tables(self):
        if not self.new_engine:
            return

        table_names = self.get_table_names()
        for table_name in table_names:
            table = ReflectedTable(self.new_engine, self.new_connection)
            table.truncate_table(table_name)

    def get_table_names(self):
        table_names = [
            "credentials",
            "sourcecategories",
            "sourcesubcategories",
            "sourcedatamodel",
            "userconfig",
            "configurationentry",
            "linkdatamodel",
            "domains",
            "usertags",
            "compactedtags",
            "usercompactedtags",
            "entrycompactedtags",
            "uservotes",
            "browser",
            "entryrules",
            "dataexport",
            "gateway",
            "modelfiles",
            "readlater",
            "searchview",
            "socialdata",
            # "blockentry",
            "blockentrylist",
            "usercomments",
            "userbookmarks",
            "usersearchhistory",
            "userentrytransitionhistory",
            "userentryvisithistory",
            "user",
        ]
        return table_names

    def print_data(self, entry, data):
        """
        If we print to SQLITE we want to see progress so we display it anyway
        """
        link = data["link"]
        title = data["title"]
        page_rating_votes = data["page_rating_votes"]

        if self.output_format == "LINES" or self.output_format == "SQLITE":
            print(f"[{page_rating_votes}] {link} - {title}")
            user_tags = ReflectedEntryCompactedTags(self.engine, self.connection)
            tags = user_tags.get_tags_string(entry.id)
            if tags:
                print(f"{tags}")
        elif self.output_format == "JSON":
            user_tags = ReflectedEntryCompactedTags(self.engine, self.connection)
            tags = user_tags.get_tags(entry.id)
            print(
                f"""
            \{ "title" : "{title}",
              "link" : "{link}",
              "page_rating_votes : {page_rating_votes},
              "tags" : {tags}
            \}"""
            )
        else:
            print("Unsupported output format")


def parse():
    parser = argparse.ArgumentParser(description="Data analyzer program")
    parser.add_argument("--db", default="catalog.db", help="DB to be scanned")
    parser.add_argument("--output-db", help="File to be created")
    parser.add_argument("--update-rss",action="store_true", help="Reads RSS to check it's title and properties")
    parser.add_argument("--read-internet-links",action="store_true", help="Reads entries to check if contains RSS. Without it only calculated RSS are returned")
    parser.add_argument(
        "--output-format",
        default="LINES",
        help="format of display. LINES, JSON, SQLITE",
    )
    parser.add_argument("--crawling-server", default="", help="Remote crawling server")

    args = parser.parse_args()

    return parser, args


def main():
    parser, args = parse()

    path = Path(args.db)
    if not path.exists():
        print("File {} does not exist".format(path))
        return

    reader = Db2Feeds(
        input_db=args.db,
        remote_server=args.crawling_server,
        output_db=args.output_db,
        output_format=args.output_format,
    )
    reader.convert()


if __name__ == "__main__":
    main()



"""
Filters out redundant things from database.
Normally for views, analysis you do not need no temporary tables.
"""

import os
import sys
import json
import time
import shutil
from pathlib import Path
import argparse

from sqlalchemy import create_engine
from .utils.reflected import *


class DbFilter(object):
    """
    Filter class
    """

    def __init__(self, input_db, output_db):
        self.input_db = input_db
        self.output_db = output_db
        self.engine = None
        self.connection = None
        self.setup()

    def setup(self):
        path = Path(self.input_db)
        if not path.exists():
            print("File {} does not exist".format(path))
            return

        new_path = Path(self.output_db)
        if new_path.exists():
            new_path.unlink()

        shutil.copy(self.input_db, self.output_db)

        self.engine = create_engine(f"sqlite:///{self.output_db}")
        self.connection = self.engine.connect()

    def is_valid(self) -> bool:
        if not self.engine:
            return False
        return True

    def close(self):
        if self.connection:
            self.connection.close()
            self.connection = None

    def truncate(self):
        table = ReflectedEntryTable(self.engine, self.connection)

        table.truncate_table("userentrytransitionhistory")
        table.truncate_table("userentryvisithistory")
        table.truncate_table("usersearchhistory")
        table.truncate_table("uservotes")
        table.truncate_table("usercompactedtags")
        table.truncate_table("usercomments")
        table.truncate_table("userbookmarks")
        table.truncate_table("user")
        table.truncate_table("userconfig")
        table.truncate_table("sourcedatamodel")
        table.truncate_table("sourcecategories")
        table.truncate_table("sourcesubcategories")
        table.truncate_table("readlater")
        table.truncate_table("modelfiles")
        table.truncate_table("gateway")
        table.truncate_table("entryrules")
        table.truncate_table("domains")
        table.truncate_table("dataexport")
        table.truncate_table("configurationentry")
        table.truncate_table("compactedtags")
        table.truncate_table("blockentrylist")

    def filter(self, conditions):
        table = ReflectedEntryTable(self.engine, self.connection)

        sql_text = f"DELETE FROM linkdatamodel WHERE {conditions};"
        # TODO delete depnded things
        table.run_sql(sql_text)
        table.vacuum()
        table.close()

    def filter_bookmarks(self):
        table = ReflectedEntryTable(self.engine, self.connection)

        sql_text = f"DELETE FROM linkdatamodel WHERE bookmarked=False;"
        # TODO delete depnded things
        table.run_sql(sql_text)
        table.vacuum()
        table.close()

    def filter_votes(self):
        table = ReflectedEntryTable(self.engine, self.connection)

        sql_text = f"DELETE FROM linkdatamodel WHERE page_rating_votes=0;"
        table.run_sql(sql_text)
        table.vacuum()
        table.close()

    def filter_redundant(self):
        """
        Not bookmarked AND without votes are redundant
        """
        table = ReflectedEntryTable(self.engine, self.connection)

        sql_text = f"DELETE FROM linkdatamodel WHERE bookmarked=False AND page_rating_votes=0;"
        table.run_sql(sql_text)
        table.vacuum()
        table.close()


def parse():
    parser = argparse.ArgumentParser(description="Data analyzer program")
    parser.add_argument("--db", default="places.db", help="DB to be scanned")
    parser.add_argument("--output-db", default="new.db", help="DB to be created")
    parser.add_argument("--bookmarked", action="store_true", help="export bookmarks")
    parser.add_argument("--votes", action="store_true", help="export if votes is > 0")
    parser.add_argument("--clean", action="store_true", help="cleans db from tables")
    parser.add_argument("-v", "--verbosity", help="Verbosity level")

    args = parser.parse_args()

    return parser, args


def main():
    start_time = time.time()
    parser, args = parse()

    thefilter = DbFilter(args.db, args.output_db)
    if not thefilter.is_valid():
        return

    thefilter.truncate()
    if args.bookmarked:
        thefilter.filter_bookmarks()
    if args.votes:
        thefilter.filter_votes()

    thefilter.vacuum()
    thefilter.close()

    end_time = time.time()
    print(f"Done in {end_time}")


if __name__ == "__main__":
    main()



def get_tables(self):
    table_names = [
        "credentials",
        "sourcecategories",
        "sourcesubcategories",
        "sourcedatamodel",
        "userconfig",
        "configurationentry",
        "linkdatamodel",
        "domains",
        "usertags",
        "compactedtags",
        "usercompactedtags",
        "entrycompactedtags",
        "uservotes",
        "browser",
        "entryrules",
        "dataexport",
        "gateway",
        "modelfiles",
        "readlater",
        "searchview",
        "socialdata",
        "blockentry",
        "blockentrylist",
        "usercomments",
        "userbookmarks",
        "usersearchhistory",
        "userentrytransitionhistory",
        "userentryvisithistory",
    ]
    return tables


def get_backup_tables(self):
    table_names = [
        "credentials",
        "sourcecategories",
        "sourcesubcategories",
        "sourcedatamodel",
        "userconfig",
        "configurationentry",
        "linkdatamodel",
        "domains",
        "usertags",
        "compactedtags",
        "usercompactedtags",
        "entrycompactedtags",
        "uservotes",
        "browser",
        "entryrules",
        "dataexport",
        "gateway",
        "modelfiles",
        "readlater",
        "searchview",
        "socialdata",
        # "blockentry", # do not backup it, the list has to be reinitialized each time
        "blockentrylist",
        "usercomments",
        "userbookmarks",
        "usersearchhistory",
        "userentrytransitionhistory",
        "userentryvisithistory",
    ]
    return tables



from sqlalchemy import MetaData, Table, select, text, inspect, insert
from sqlalchemy import Index


class ReflectedTable(object):
    def __init__(self, engine, connection):
        self.engine = engine
        self.connection = connection

    def get_table(self, table_name):
        destination_metadata = MetaData()
        destination_table = Table(
            table_name, destination_metadata, autoload_with=self.engine
        )
        return destination_table

    def truncate_table(self, table_name):
        sql_text = f"DELETE FROM {table_name};"
        self.connection.execute(text(sql_text))
        self.connection.commit()

    def create_index(self, table, column_name):
        index_name = f"idx_{table.name}_{column_name}"
        index = Index(index_name, getattr(table.c, column_name))

        index.create(bind=self.engine)

    def vacuum(self):
        self.connection.execute(text("VACUUM"))

    def close(self):
        pass

    def insert_json_data(self, table_name, json_data: dict):
        table = self.get_table(table_name)

        stmt = insert(table).values(**json_data)
        self.connection.execute(stmt)
        self.connection.commit()

    def print_summary(self, print_columns=False):
        inspector = inspect(self.engine)
        tables = inspector.get_table_names()

        for table in tables:
            row_count = self.connection.execute(
                text(f"SELECT COUNT(*) FROM {table}")
            ).scalar()
            print(f"Table: {table}, Row count: {row_count}")

            if print_columns:
                columns = inspector.get_columns(table)
                column_names = [column["name"] for column in columns]
                print(f"Columns in {table}: {', '.join(column_names)}")

    def run_sql(self, sql_text):
        self.connection.execute(text(sql_text))
        self.connection.commit()


class ReflectedEntryTable(ReflectedTable):
    def truncate(self):
        self.truncate_table("linkdatamodel")

    def get_entries(self):
        destination_table = self.get_table("linkdatamodel")

        entries_select = select(destination_table)

        result = self.connection.execute(entries_select)
        entries = result.fetchall()

        for entry in entries:
            yield entry

    def get_entries_good(self):
        destination_table = self.get_table("linkdatamodel")

        stmt = (
            select(destination_table)
            .where(destination_table.c.page_rating_votes > 0)
            .order_by(destination_table.c.page_rating_votes.desc())
        )

        result = self.connection.execute(stmt)
        entries = result.fetchall()

        for entry in entries:
            yield entry


class ReflectedUserTags(ReflectedTable):
    def get_tags_string(self, entry_id):
        destination_table = self.get_table("usertags")

        stmt = select(destination_table).where(destination_table.c.entry_id == entry_id)

        tags = ""

        result = self.connection.execute(stmt)
        rows = result.fetchall()
        for row in rows:
            if tags:
                tags += ", "

            tags += "#" + row.tag

        return tags

    def get_tags(self, entry_id):
        destination_table = self.get_table("usertags")

        stmt = select(destination_table).where(destination_table.c.entry_id == entry_id)

        tags = []

        result = self.connection.execute(stmt)
        rows = result.fetchall()
        for row in rows:
            tags.append(row.tag)

        return tags


class ReflectedEntryCompactedTags(ReflectedTable):
    def get_tags_string(self, entry_id):
        destination_table = self.get_table("entrycompactedtags")

        stmt = select(destination_table).where(destination_table.c.entry_id == entry_id)

        tags = ""

        result = self.connection.execute(stmt)
        rows = result.fetchall()
        for row in rows:
            if tags:
                tags += ", "

            tags += "#" + row.tag

        return tags

    def get_tags(self, entry_id):
        destination_table = self.get_table("entrycompactedtags")

        stmt = select(destination_table).where(destination_table.c.entry_id == entry_id)

        tags = []

        result = self.connection.execute(stmt)
        rows = result.fetchall()
        for row in rows:
            tags.append(row.tag)

        return tags


class ReflectedSourceTable(ReflectedTable):
    def truncate(self):
        self.truncate_table("sourcedatamodel")

    def get_source(self, source_id):
        destination_table = self.get_table("sourcedatamodel")

        stmt = select(destination_table).where(destination_table.c.id == source_id)

        result = self.connection.execute(stmt)
        return result.first()


class ReflectedSocialData(ReflectedTable):
    def truncate(self):
        self.truncate_table("socialdata")

    def get(self, entry_id):
        destination_table = self.get_table("socialdata")

        stmt = select(destination_table).where(destination_table.c.entry_id == entry_id)

        result = self.connection.execute(stmt)
        return result.first()



"""
Converts database to JSON.
"""

import os
import sys
import json
import shutil
from pathlib import Path
import argparse

from sqlalchemy import create_engine
from .utils.reflected import *


class Db2JSON(object):

    def __init__(self, input_db, output_dir, format=None, rows_max=1000):
        self.input_db = input_db
        self.output_dir = output_dir

        self.format = format
        self.rows_max = rows_max

        self.file_index = 0
        self.entry_index = 0
        self.handle = None

        self.rows = []

        self.processed = 0
        self.all = 0

        self.setup()

    def setup(self):
        path = Path(self.input_db)
        if not path.exists():
            print("File {} does not exist".format(path))
            return

        if self.output_dir and self.output_dir != ".":
            new_path = Path(self.output_dir)
            if new_path.exists():
                shutil.rmtree(new_path)
            new_path.mkdir()

        self.engine = create_engine(f"sqlite:///{self.input_db}")

    def write(self, entry):
        """Write entries to the specified directory, 1000 per file."""
        if self.handle == None:
            file_path = str(self.get_file_path())
            self.handle = open(file_path, "w")

        row = self.get_entry_json_data(entry)

        self.rows.append(row)

        self.entry_index += 1

        sys.stdout.write(f"{self.file_index}/{self.entry_index:04d}\r")

        if self.entry_index == self.rows_max:
            self.file_index += 1
            self.entry_index = 0
            self.finish_stream()

            file_path = str(self.get_file_path())
            self.handle = open(file_path, "w")

    def get_entry_json_data(self, entry):
        date_published = entry.date_published
        if date_published:
            date_published = date_published.isoformat()

        date_dead_since = entry.date_dead_since
        if date_dead_since:
            date_dead_since = date_dead_since.isoformat()

        row = {
            "link": entry.link,
            "description": entry.description,
            "author": entry.author,
            "album": entry.album,
            "bookmarked": entry.bookmarked,
            "date_dead_since": date_dead_since,
            "date_published": date_published,
            "language": entry.language,
            "manual_status_code": entry.manual_status_code,
            "page_rating": entry.page_rating,
            "page_rating_contents": entry.page_rating_contents,
            "page_rating_votes": entry.page_rating_votes,
            "page_rating_visits": entry.page_rating_visits,
            "permanent": entry.permanent,
            "source_url": entry.source_url,
            "status_code": entry.status_code,
            "thumbnail": entry.thumbnail,
            "title": entry.title,
            "age": entry.age,
            "id": entry.id,
        }

        social_table = ReflectedSocialData(self.engine, self.connection)
        social_data = social_table.get(entry.id)
        if social_data:
            row.setdefault("thumbs_up", social_data.thumbs_up)
            row.setdefault("thumbs_down", social_data.thumbs_down)
            row.setdefault("view_count", social_data.view_count)
            row.setdefault("rating", social_data.rating)
            row.setdefault("upvote_ratio", social_data.upvote_ratio)
            row.setdefault("upvote_diff", social_data.upvote_diff)
            row.setdefault("upvote_view_ratio", social_data.upvote_view_ratio)
            row.setdefault("stars", social_data.stars)
            row.setdefault("followers_count", social_data.followers_count)

        tags_table = ReflectedUserTags(self.engine, self.connection)
        tags = tags_table.get_tags(entry.id)
        row["tags"] = tags

        return row

    def get_file_path(self):
        filename = "{}_{}.json".format(self.format, str(self.file_index))
        if self.output_dir and self.output_dir != ".":
            return Path(self.output_dir) / filename
        else:
            return Path(filename)

    def close(self):
        if self.handle:
            self.finish_stream()
        self.handle = None

    def finish_stream(self):
        if not self.handle:
            return

        try:
            string = json.dumps(self.rows, indent=4)
            self.handle.write(string)
        except ValueError as e:
            print(f"Error writing file {file_path}: {e}")
        self.handle.close()
        self.rows = []

    def convert(self):
        with self.engine.connect() as connection:
            self.connection = connection
            table = ReflectedEntryTable(self.engine, connection)

            for entry in table.get_entries():
                # print(entry)
                self.write(entry)


def parse():
    parser = argparse.ArgumentParser(description="Data analyzer program")
    parser.add_argument("--db", default="places.db", help="DB to be scanned")
    parser.add_argument("--output-dir", default="json", help="Output directory")
    parser.add_argument(
        "--rows-max", default=1000, action="store_true", help="Number of rows per file"
    )
    parser.add_argument("-f", "--format", default="entries", help="file name format")
    parser.add_argument("-v", "--verbosity", help="Verbosity level")

    args = parser.parse_args()

    return parser, args


def main():
    parser, args = parse()

    f = Db2JSON(
        input_db=args.db,
        output_dir=args.output_dir,
        format=args.format,
        rows_max=args.rows_max,
    )
    f.convert()
    f.close()


if __name__ == "__main__":
    main()



from sqlalchemy import and_, or_, not_, func, MetaData, Table, select

from .omnisearch import (
    SingleSymbolEvaluator,
    EquationEvaluator,
    OmniSearch,
)


class AlchemySymbolEvaluator(SingleSymbolEvaluator):
    """
    return 1 if true
    """

    def __init__(self, table, ignore_case=False):
        self.table = table
        self.ignore_case = ignore_case

    def evaluate_complex_symbol(self, symbol, condition_data):
        # TODO make todo check if symbol exists in table?

        if condition_data[1] == "==":
            if self.ignore_case:
                column = self.table.c[condition_data[0]]
                return func.lower(column) == condition_data[2].lower()
            else:
                return self.table.c[condition_data[0]] == condition_data[2]

        if condition_data[1] == "!=":
            if self.ignore_case:
                column = self.table.c[condition_data[0]]
                return func.lower(column) != condition_data[2].lower()
            else:
                return self.table.c[condition_data[0]] != condition_data[2]

        if condition_data[1] == ">":
            return self.table.c[condition_data[0]] > condition_data[2]

        if condition_data[1] == "<":
            return self.table.c[condition_data[0]] < condition_data[2]

        if condition_data[1] == ">=":
            return self.table.c[condition_data[0]] >= condition_data[2]

        if condition_data[1] == "<=":
            return self.table.c[condition_data[0]] <= condition_data[2]

        if condition_data[1] == "=":
            symbol = condition_data[2]
            symbol = symbol.replace("*", "%")

            if self.ignore_case:
                return self.table.c[condition_data[0]].ilike(symbol)
            else:
                return self.table.c[condition_data[0]].like(symbol)

        raise IOError("Unsupported operator")

    def evaluate_simple_symbol(self, symbol):
        """
        TODO we could check by default if entry link == symbol, or sth
        """
        if self.ignore_case:
            symbol = symbol.replace("*", "%")
            return or_(
                self.table.c["link"].ilike(symbol),
                self.table.c["title"].ilike(symbol),
                self.table.c["description"].ilike(symbol),
            )
        else:
            symbol = symbol.replace("*", "%")
            return or_(
                self.table.c["link"].like(symbol),
                self.table.c["title"].like(symbol),
                self.table.c["description"].like(symbol),
            )


class AlchemyEquationEvaluator(EquationEvaluator):
    def evaluate_function(self, operation_symbol, function, args0, args1):
        if function == "And":  # & sign
            return and_(args0, args1)
        elif function == "Or":  # | sign
            return or_(args0, args1)
        elif function == "Not":  # ~ sign
            return not_(args0)
        else:
            raise NotImplementedError("Not implemented function: {}".format(function))


class AlchemyRowHandler(object):
    def handle_row(self, row):
        pass


class AlchemySearch(object):
    def __init__(self, db, search_term, row_handler=None, args=None, connection=None):
        self.db = db
        self.connection = connection
        self.search_term = search_term
        self.alchemy_row_handler = row_handler

        self.args = args

        self.get_destination_table()

    def search(self):
        rows = self.get_filtered_objects()

        for row in rows:
            self.alchemy_row_handler.handle_row(row)

    def get_destination_table(self):
        destination_metadata = MetaData()

        if self.args and "table" in self.args:
            self.destination_table = Table(
                self.args.table, destination_metadata, autoload_with=self.db
            )
        else:
            self.destination_table = Table(
                "linkdatamodel", destination_metadata, autoload_with=self.db
            )

    def get_query_conditions(self):
        ignore_case = False
        if self.args and self.args.ignore_case:
            ignore_case = True

        symbol_evaluator = AlchemySymbolEvaluator(self.destination_table, ignore_case)
        equation_evaluator = AlchemyEquationEvaluator(
            self.search_term, symbol_evaluator
        )

        search = OmniSearch(self.search_term, equation_evaluator=equation_evaluator)
        combined_query_conditions = search.get_combined_query()
        return combined_query_conditions

    def get_filtered_objects(self):
        combined_query_conditions = self.get_query_conditions()

        rows = []

        order_by_column_name = "id"
        if self.args and self.args.order_by:
            order_by_column_name = self.args.order_by

        order_by_column = getattr(self.destination_table.c, order_by_column_name, None)

        if order_by_column is None:
            raise AttributeError(f"Invalid order_by column: {self.args.order_by}")

        if self.args:
            # Determine sorting order
            order_by_clause = (
                order_by_column.asc()
                if self.args.asc
                else order_by_column.desc() if self.args.desc else order_by_column.asc()
            )
        else:
            order_by_clause = order_by_column.asc()

        # Use select() for SQLAlchemy Core
        stmt = (
            select(self.destination_table)
            .where(combined_query_conditions)
            .order_by(order_by_clause)
        )

        # Execute the query
        result = self.connection.execute(stmt)

        # Fetch all results
        rows = result.fetchall()

        return rows
